----------------------------------------
-- copy a table and all its subtables --
----------------------------------------
local function tcopy(intable,outtable,overwrite,written)
	--written is an internal table to track references to already copied tables and keep those intact
	--create outtable if it doesn't exist
	outtable = outtable or {}
	--create written if this is the first call (user call)
	written = written or {}
	--add copied table to written
	written[tostring(intable)] = outtable
	--loop through data
	for key, data in pairs(intable) do
		--if overwrite or not exists, write value
		if (not outtable[key]) or overwrite then
			if type(data) == "table" then
				outtable[key] = written[tostring(data)] or tcopy(data,nil,nil,written) -- we don't pass overwrite because the table copied to is freshly created anyways
			else
				outtable[key] = data
			end
		end
	end
	return outtable
end
local _, addon_data = ...
local db = addon_data.db
if not db then return end
local temp = addon_data.templ
local function resolveInherits(t, path, aboveInherit)
	path = path or "" -- root table
	local inherit = t.inherits
	if inherit or (inherit == nil and aboveInherit) then
		inherit = type(inherit) == "string" and inherit or aboveInherit
		local inheritedTable = temp[inherit]
		if not inheritedTable then
			print("|cffff0000InspectDetail error:|r Invalid inherit "..inherit.." (No template with that name exists)")
			return
		end
		for key in string.gmatch(path, "[^|]+") do
			inheritedTable = inheritedTable[key]
			if not inheritedTable then
				print("|cffff0000InspectDetail error:|r Invalid inheritance path "..path.." in inheritance "..inherit.." (That sub-table does not exist).")
				return
			end
		end
		resolveInherits(inheritedTable)
		tcopy(inheritedTable, t, false)
	end
	for key, value in pairs(t) do
		if type(value) == "table" then
			resolveInherits(value, path.."|"..key, inherit or aboveInherit)
		end
	end
	t.inherits = nil
	local cnt=2
	while t["inherits"..cnt] do
		t["inherits"..(cnt-1)]=t["inherits"..cnt]
		t["inherits"..cnt] = nil
		cnt=cnt+1
	end
	if cnt > 2 then
		resolveInherits(t,  path, aboveInherit)
	end
end
local fetchtable
local function fetchstats(itemLink)
	fetchtable = fetchtable and table.wipe(fetchtable) or {}
	GetItemStats(itemLink, fetchtable)
	return fetchtable
end
local addon = CreateFrame("Frame", "InspectDetail")
addon.needsUpdate = {}
local iconed_frames = {}
local worktable = {
	question={
		gems={},
		extramessages={},
		enchants={},
		stats={},
	},
	wrong={
		gems={},
		extramessages={},
		enchants={},
		stats={},
	},
	strings={
		wrong={},
		question={},
	},
}
local function showtooltip_q(self)
	local text = self.text
	if not text then return end
	GameTooltip:ClearLines()
	GameTooltip:SetOwner(self, "ANCHOR_NONE")
	GameTooltip:SetPoint("TOP", self:GetParent(), "BOTTOM")
	GameTooltip:AddLine("Questionable on this item:", 1, 0.82, 0)
	for line in string.gmatch(text, "[^\n]+") do
		GameTooltip:AddLine(line, 1, 1, 1)
	end
	GameTooltip:Show()
end
local function showtooltip_w(self)
	local text = self.text
	if not text then return end
	GameTooltip:ClearAllPoints()
	GameTooltip:ClearLines()
	GameTooltip:SetOwner(self, "ANCHOR_NONE")
	GameTooltip:SetPoint("TOP", self:GetParent(), "BOTTOM")
	GameTooltip:AddLine("Wrong on this item:", 1, 0.82, 0)
	for line in string.gmatch(text, "[^\n]+") do
		GameTooltip:AddLine(line, 1, 1, 1)
	end
	GameTooltip:Show()
end
local hidetooltip = function() GameTooltip:Hide() end
local function doScan(t,blizzButton,spec)
	t.wrong:Hide()
	t.questionable:Hide()
	local unit = InspectFrame.unit
	local class = (select(2, UnitClass(unit))) or InspectFrame.cacheClass
	InspectFrame.cacheClass = class
	if not class then
		return
	end
	local ratingTable = db[class][spec]
	if not ratingTable then
		print("|cffff0000InspectDetail error:|r No rating table found for: ", class, spec)
		return
	end
	resolveInherits(ratingTable)
	local itemLink = GetInventoryItemLink(unit, t.id)
	if not itemLink then return end
	local itemID = tonumber(string.match(itemLink, "|Hitem:(%d+):"))
	local slotName = blizzButton:GetName():match("Inspect(.+)Slot")
	local _,blankLink = GetItemInfo(itemID)
	local stats = fetchstats(blankLink)
	local num_gems=(stats.EMPTY_SOCKET_RED or 0)+(stats.EMPTY_SOCKET_YELLOW or 0)+(stats.EMPTY_SOCKET_BLUE or 0)+(stats.EMPTY_SOCKET_PRISMATIC or 0)+(stats.EMPTY_SOCKET_HYDRAULIC or 0)+(stats.EMPTY_SOCKET_COGWHEEL or 0)
	local isQuestion, isWrong=false,false
	for _, tbl in pairs(worktable) do
		for key in pairs(tbl) do
			table.wipe(tbl[key])
		end
	end
	for stat, amount in pairs(stats) do
		local statRating = ratingTable.stats[stat] or ratingTable.stats.noclass
		if statRating == 1 then
			tinsert(worktable.question.stats, stat)
			isQuestion=true
		elseif statRating == 0 then
			tinsert(worktable.wrong.stats, stat)
			isWrong=true
		end
	end
	--empty slot check
	for slotID=1, (slotName == "Waist") and (num_gems+1) or num_gems do
		local gemLink=select(2,GetItemGem(itemLink,slotID))
		if gemLink then
			local gemID=tonumber(gemLink:match("|Hitem:(%d+):"))
			local gemRating=ratingTable.gemming[gemID] or ratingTable.gemming.noclass
			if type(gemRating)=="string" then
				-- temporary!
				gemRating=2
			end
			if gemRating == 1 then
				tinsert(worktable.question.gems, gemID)
				isQuestion=true
			elseif gemRating == 0 then
				tinsert(worktable.wrong.gems, gemID)
				isWrong=true
			end
		else
			if slotID > num_gems then
				worktable.wrong.extramessages.missing_buckle = true
			else
				worktable.wrong.extramessages.missing_gems=(worktable.wrong.extramessages.missing_gems or 0)+1
			end
			isWrong = true
		end
	end
	--enchant check
	local enchantID = tonumber(itemLink:match("|Hitem:%d+:(%d+):"))
	if enchantID > 0 then
		-- is enchanted
		local enchantRating = ratingTable.enchanting[enchantID] or ratingTable.enchanting.noclass
		if type(enchantRating) == "string" then
			-- temporary again!
			enchantRating=2
		end
		if enchantRating == 1 then
			tinsert(worktable.question.enchants, enchantID) -- enchant link should be placed here later on
			isQuestion=true
		elseif enchantRating == 0 then
			tinsert(worktable.wrong.enchants, enchantID) -- same as above
			isWrong=true
		end
	else
		-- isn't enchanted
		local noRating = ratingTable.noenchant[slotName] or ratingTable.noenchant.noclass
		if type(noRating) == "string" then
			-- temporary!
			noRating = 2
		end
		if noRating == 1 then
			tinsert(worktable.question.enchants, "missing")
			isQuestion=true
		elseif noRating == 0 then
			tinsert(worktable.wrong.enchants, "missing")
			isWrong=true
		end
	end
	if isWrong then
		local s = worktable.strings.wrong
		for _, stat in ipairs(worktable.wrong.stats) do
			tinsert(s, "Item has ".._G[stat])
		end
		for _, gemID in ipairs(worktable.wrong.gems) do
			tinsert(s, "Item is socketed with "..(select(2,GetItemInfo(gemID))))
		end
		for _, enchant in ipairs(worktable.wrong.enchants) do
			tinsert(s, "Enchant is "..enchant)
		end
		for msg, value in pairs(worktable.wrong.extramessages) do
			if msg == "missing_gems" then
				tinsert(s, value.." |4missing gem:missing gems")
			elseif msg == "missing_buckle" then
				tinsert(s, "Missing belt buckle")
			end
		end
		t.wrong.text = table.concat(s, "\n")
		t.wrong:Show()
	end
	if isQuestion then
		local s = worktable.strings.question
		for _, stat in ipairs(worktable.question.stats) do
			tinsert(s, "Item has ".._G[stat])
		end
		for _, gemID in ipairs(worktable.question.gems) do
			tinsert(s, "Item is socketed with "..(select(2, GetItemInfo(gemID))))
		end
		for _, enchant in ipairs(worktable.question.enchants) do
			tinsert(s, "Enchant is "..enchant)
		end
		for msg, value in pairs(worktable.question.extramessages) do
		end
		t.questionable.text = table.concat(s, "\n")
		t.questionable:Show()
	end
end
local function ourHook(blizzButton)
	local blizzName = blizzButton:GetName()
	local spec = (GetPrimaryTalentTree(true, false, nil))
	if not spec then
		tinsert(addon.needsUpdate, blizzName)
		addon:RegisterEvent("INSPECT_TALENT_READY")
	end
	local t = iconed_frames[blizzName]
	if not t then
		t = {id=blizzButton:GetID()}
		t.questionable = CreateFrame("Frame", blizzName.."QuestionableIcon", blizzButton)
		t.questionable:SetSize(19.2,24)
		t.questionable:SetPoint("TOPLEFT", -3, 3)
		t.questionable.texture = t.questionable:CreateTexture()
		t.questionable.texture:SetTexture("Interface\\WorldMap\\UI-WorldMap-QuestIcon")
		t.questionable.texture:SetTexCoord(0.13,0.41,0.07,0.41)
		t.questionable.texture:SetAllPoints()
		--t.questionable:SetNormalTexture(t.questionable.texture)
		t.questionable.texture2 = t.questionable:CreateTexture(nil,"HIGHLIGHT")
		t.questionable.texture2:SetTexture("Interface\\WorldMap\\UI-WorldMap-QuestIcon")
		t.questionable.texture2:SetTexCoord(0.63,0.91,0.07,0.41)
		t.questionable.texture2:SetAllPoints()
		t.questionable:EnableMouse(true)
		t.questionable:SetScript("OnEnter",showtooltip_q)
		t.questionable:SetScript("OnLeave",hidetooltip)
		t.wrong = CreateFrame("Frame", blizzName.."WrongIcon", blizzButton)
		t.wrong:SetSize(19.2,19.2)
		t.wrong:SetPoint("TOPRIGHT", 3, 3)
		t.wrong.texture = t.wrong:CreateTexture()
		t.wrong.texture:SetTexture("Interface\\AddOns\\InspectDetail\\fail")
		t.wrong.texture:SetAllPoints()
		t.wrong.texture:SetVertexColor(0.85,0.85,0.85)
		t.wrong.texture2 = t.wrong:CreateTexture(nil, "HIGHLIGHT")
		t.wrong.texture2:SetTexture("Interface\\AddOns\\InspectDetail\\fail")
		t.wrong.texture2:SetAllPoints()
		t.wrong:EnableMouse(true)
		t.wrong:SetScript("OnEnter",showtooltip_w)
		t.wrong:SetScript("OnLeave",hidetooltip)
		t.onupdate = CreateFrame("Frame", nil, blizzButton)
		t.onupdate:SetScript("OnUpdate", function(self,elapsed)
			self.elapsed = (self.elapsed or 0)+elapsed
			if self.elapsed < 0.5 then return end
			self.elapsed = 0
			local link = GetInventoryItemLink(InspectFrame.unit, t.id)
			if (not self.curLink) or ((self.curLink ~= link) and link) then
				doScan(t, blizzButton, spec)
			end
		end)
		iconed_frames[blizzName] = t
	end
end
addon:SetScript("OnEvent", function(self, event, name)
	if event == "ADDON_LOADED" then
		if name == "Blizzard_InspectUI" then
			hooksecurefunc("InspectPaperDollItemSlotButton_Update", ourHook)
			self:UnregisterEvent("ADDON_LOADED")
		end
	end
	if event == "INSPECT_TALENT_READY" then
		for _, f in ipairs(self.needsUpdate) do
			if _G[f] then
				ourHook(_G[f])
			end
		end
		self:UnregisterEvent("INSPECT_TALENT_READY")
		table.wipe(self.needsUpdate)
	end
end)
addon:RegisterEvent("ADDON_LOADED")
-- replace default inspect slash command with a uID-recognizing one
SlashCmdList["UNITID_INSPECT"]=function(input)
	while input:sub(0,1) == " " do
		input = input:sub(2)
	end
	while input:sub(-1) == " " do
		input = input:sub(0,strlen(input)-1)
	end
	InspectUnit(UnitExists(input) and input or "target")
end
local sCnt=1
local slash_string = "SLASH_INSPECT"..sCnt
while _G[slash_string] do
	_G["SLASH_UNITID_INSPECT"..sCnt]=_G[slash_string]
	_G[slash_string]=nil
	sCnt=sCnt+1
	slash_string = "SLASH_INSPECT"..sCnt
end
